/*
 * Copyright (c) 2012 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Thu Aug  2 2012
 */
#include <sys/types.h>
#include <assert.h>
#include <err.h>
#include <fcntl.h>
#include <getopt.h>
#include <pthread.h>
#include <unistd.h>

#include <cerrno>

#include "ros/init.h"
#include "ros/master.h"
#include "ros/callback_queue.h"

#include "genom3/c/client.h"
#include "private.h"


/* --- local data ---------------------------------------------------------- */

#define ros_cbevent	'e'
#define ros_end		'q'

static int ros_nodeinit = 0;
static int ros_evpipe[2] = { -1, -1 };
static pthread_t ros_evtid;
static int ros_evtid_init = 0;

static void *	ros_evtask(void *data);


/* --- ros_init ------------------------------------------------------------ */

int
ros_init(int argc, char *argv[], char ebuffer[256], size_t esize)
{
  int s;

  /* global ros init */
  if (!ros_nodeinit) {
    try {
      ros::init(argc, argv, "genom_c_client",
                ros::init_options::NoSigintHandler |
                ros::init_options::AnonymousName |
                ros::init_options::NoRosout);
      if (!ros::master::check()) {
        snprintf(ebuffer, esize, "roscore may not be running on %s:%u",
                 ros::master::getHost().c_str(), ros::master::getPort());
        return GENOM_CLIENT_ERR_FLAG;
      }
      ros::start();
    }
    catch (ros::Exception e) {
      snprintf(ebuffer, esize, "%s", e.what());
      return GENOM_CLIENT_ERR_FLAG;
    }

    if (pipe(ros_evpipe) ||
        fcntl(ros_evpipe[0], F_SETFL, O_NONBLOCK))
      goto error;
  }
  ros_nodeinit++;

  /* create event task */
  if (!ros_evtid_init) {
    s = pthread_create(&ros_evtid, NULL, ros_evtask, NULL);
    if (s) goto error;
    ros_evtid_init = 1;
  }

  return 0;

error:
  return GENOM_SYS_ERR_FLAG | errno;
}


/* --- ros_fini ------------------------------------------------------------ */

void
ros_fini(void)
{
  int i;

  if (ros_nodeinit) ros_nodeinit--;
  if (ros_nodeinit) return;

  ros::shutdown();

  if (ros_evtid_init) {
    pthread_join(ros_evtid, NULL);
    ros_evtid_init = 0;
  }

  for(i=0; i<2; i++) {
    if (ros_evpipe[i] > 0) {
      close(ros_evpipe[i]);
      ros_evpipe[i] = -1;
    }
  }
}


/* --- ros_getopt ---------------------------------------------------------- */

/* options */
static const char shortopts_string[] = "+i:";
static struct option longopts_list[] = {
  { "name",		required_argument,	NULL,	'i'},
  { NULL, 0, NULL, 0}
};

int
ros_getopt(int argc, char *argv[], struct ros_options *opts)
{
  int c;

  opts->instance = NULL;

  /* reset getopt */
#ifdef HAVE_OPTRESET
  optreset = 1;
#endif
#ifdef __GLIBC__
  optind = 0; /* not "1", see gnu manual for getopt(3) ... */
#else
  optind = 1;
#endif

  /* parse command line options */
  opterr = 0;
  while (
    (c = getopt_long(argc, argv, shortopts_string, longopts_list, NULL)) != -1)
    switch (c) {
      case 0: break;

      case 'i': opts->instance = optarg; break;

      case '?':
      default:
        c = '?'; break;
    }
  if (argc > optind) c = '?';
  return (c == '?') ? EINVAL : 0;
}


/* --- ros_newrqst --------------------------------------------------------- */

int
ros_newrqst(genom_client h, size_t outsize, int *rqstid)
{
  genom_client_s::genom_client_request *a;
  int i;
  assert(h);

  *rqstid = -1;
  for(i = h->next; i < h->next + h->nactive; i++)
    if (h->active_rqst[i%h->nactive].status == ROS_CLIENT_FREE) {
      *rqstid = i%h->nactive;
      break;
    }
  if (*rqstid < 0) {
    a = (genom_client_s::genom_client_request *)
      realloc(h->active_rqst, (h->nactive+1)*sizeof(*a));
    if (!a) return GENOM_SYS_ERR_FLAG | errno;

    *rqstid = h->nactive;
    h->nactive++;
    h->active_rqst = a;
  }
  h->next = *rqstid + 1;

  h->active_rqst[*rqstid].info = NULL;
  h->active_rqst[*rqstid].aid = -1;
  h->active_rqst[*rqstid].status = ROS_CLIENT_FREE;
  h->active_rqst[*rqstid].pstatus = ROS_CLIENT_FREE;
  h->active_rqst[*rqstid].error = 0;
  h->active_rqst[*rqstid].output = malloc(outsize);
  h->active_rqst[*rqstid].sentcb = NULL;
  h->active_rqst[*rqstid].donecb = NULL;
  h->active_rqst[*rqstid].cb_data = NULL;
  if (!h->active_rqst[*rqstid].output) return GENOM_SYS_ERR_FLAG | errno;

  return 0;
}


/* --- ros_eventfd --------------------------------------------------------- */

int
ros_eventfd(void)
{
  return ros_evpipe[0];
}


/* --- ros_service_info ---------------------------------------------------- */

const struct genom_service_info *
ros_service_info(genom_client h, int rqstid)
{
  if (rqstid < 0 || rqstid >= h->nactive) return NULL;
  if (h->active_rqst[rqstid].status == ROS_CLIENT_FREE) return NULL;
  return h->active_rqst[rqstid].info;
}


/* --- ros_waitevent ------------------------------------------------------- */

int
ros_waitevent(void)
{
  fd_set fdset;
  int s;

  printf("ros_waitevent\n");

  FD_ZERO(&fdset);
  do {
    FD_SET(ros_evpipe[0], &fdset);
    s = select(ros_evpipe[0]+1, &fdset, NULL, NULL, NULL);
  } while(s < 0 && errno == EINTR);

  return s < 0 ? errno : 0;
}


/* --- ros_waitack --------------------------------------------------------- */

int
ros_waitack(genom_client h, int rqstid)
{
  int s;
  if (rqstid < 0 || rqstid >= h->nactive)
    return h->error = GENOM_SYS_ERR_FLAG | ENOENT;

  printf("ros_waitack\n");

  while(h->active_rqst[rqstid].pstatus == ROS_CLIENT_WAIT_ACK) {
    s = ros_waitevent();
    if (s) return h->error = s;

    s = ros_doevents(h);
    if (s) return h->error = s;
  }
  return 0;
}


/* --- ros_waitterm -------------------------------------------------------- */

int
ros_waitterm(genom_client h, int rqstid)
{
  int s;
  if (rqstid < 0 || rqstid >= h->nactive)
    return h->error = GENOM_SYS_ERR_FLAG | ENOENT;

  while(h->active_rqst[rqstid].pstatus == ROS_CLIENT_WAIT_ACK ||
        h->active_rqst[rqstid].status == ROS_CLIENT_WAIT_TERM) {
    s = ros_waitevent();
    if (s) return h->error = s;

    s = ros_doevents(h);
    if (s) return h->error = s;
  }
  return 0;
}


/* --- ros_done ------------------------------------------------------------ */

int
ros_done(genom_client h, int rqstid)
{
  if (rqstid < 0 || rqstid >= h->nactive) return 1;
  if (h->active_rqst[rqstid].status == ROS_CLIENT_WAIT_ACK ||
      h->active_rqst[rqstid].status == ROS_CLIENT_WAIT_TERM)
    return 0;
  return 1;
}


/* --- ros_clean ----------------------------------------------------------- */

int
ros_clean(genom_client h, int rqstid)
{
  struct genom_client_s::genom_client_request *a;
  if (rqstid < 0 || rqstid >= h->nactive)
    return h->error = GENOM_SYS_ERR_FLAG | ENOENT;

  a = &h->active_rqst[rqstid];
  if (a->status == ROS_CLIENT_FREE)
    return h->error = GENOM_SYS_ERR_FLAG | ENOENT;

  if (a->status == ROS_CLIENT_WAIT_ACK ||
      a->status == ROS_CLIENT_WAIT_TERM) {
    //    h->error = csClientRqstIdFree(h->csid, a->rid);
    if (h->error) return h->error;
  }

  a->status = ROS_CLIENT_FREE;
  if (a->output) {
    if (a->info->fini_output) a->info->fini_output(a->output);
    free(a->output);
  }
  return 0;
}


/* --- ros_doevents -------------------------------------------------------- */

int
ros_doevents(genom_client h)
{
  struct genom_client_s::genom_client_request *a;
  ssize_t r;
  char e[16];
  int i;

  if (ros_evpipe[0] < 0) return h->error = GENOM_SYS_ERR_FLAG | EBADF;

  do {
    do {
      r = read(ros_evpipe[0], e, sizeof(e));
    } while (r == -1 && errno == EINTR);
    if (r == 0) return h->error = GENOM_SYS_ERR_FLAG | EBADF;

    if (r == -1) {
      if ((errno == EAGAIN || errno == EWOULDBLOCK)) break;
      return h->error = GENOM_SYS_ERR_FLAG | errno;
    }

    for(i=0; i<r; i++) {
      switch(e[i]) {
        case ros_end:
          return h->error = GENOM_SYS_ERR_FLAG | ENODEV;

        case ros_cbevent:
          break;

        default: assert(!"reached unreachable");
      }
    }
  } while(r == sizeof(e));

  for(i = 0; i < h->nactive; i++) {
    a = &h->active_rqst[i];
    if (a->status == a->pstatus) continue;

    switch(a->status) {
      case ROS_CLIENT_WAIT_TERM:
        a->pstatus = a->status;
        h->error = a->error;
        if (a->sentcb) {
          if (a->error) {
            if ((a->error & 0xffff0000) != GENOM_COMP_ERR_FLAG)
              (*a->sentcb)(h, i, a->error, 0, NULL, a->cb_data);
            else
              (*a->sentcb)(h, i, 0, a->error & 0xffff, NULL, a->cb_data);
          } else
            (*a->sentcb)(h, i, 0, 0, NULL, a->cb_data);
        }
        break;

      case ROS_CLIENT_TERM:
        a->pstatus = a->status;
        h->error = a->error;
        if (a->donecb) {
          if (a->error) {
            if ((a->error & 0xffff0000) != GENOM_COMP_ERR_FLAG)
              (*a->donecb)(h, i, a->error, 0, NULL, a->cb_data);
            else
              (*a->donecb)(h, i, 0, a->error & 0xffff, NULL, a->cb_data);
          } else
            (*a->donecb)(h, i, 0, 0, a->output, a->cb_data);
        }
        break;
    }
  }

  return 0;
}


/* --- pocolibs_evtask ----------------------------------------------------- */

static void *
ros_evtask(void *data)
{
  ros::CallbackQueue *queue = ros::getGlobalCallbackQueue();
  char c;

  c = ros_cbevent;
  while(ros::ok()) {
    queue->callOne(ros::WallDuration(1.));
    if (!ros_nodeinit) break;
    write(ros_evpipe[1], &c, 1);
  }

  c = ros_end;
  write(ros_evpipe[1], &c, 1);
  return NULL;
}
